"""
Cleans and formats HMET time series data into a GSSHA HMET file, adding missing timesteps and filling missing values
with the GSSHA default values. All input gage files must be tab delimited and are assumed to be YYYY\tMM\tDD\tHH\tMM\tValue 
format.

Aligns precipitation data and outputs a GSSHA file with a specified event spacing. All input gage files must be tab
delimited and are assumed to be YYYY\tMM\tDD\tHH\tStation Pressure (in)\t Relative Humidity(%)\t Cloud (%)\t WindSpeed (kts)\t
Temperature (F)\t Direct Radiation (W/M*M)\t Total Radiation(W/M*M) format. The user must format the columns and perform
any quality control, preprocessing, or unit conversion. Only hourly input data is handled, with an hourly output series.
The timestamp can be converted from the local time into GMT by providing a local offset from GMT.

To use this script, the user should modify the entries in the user inputs section. Below are the steps for using this
script:

1) Modify the name of the target input file
2) Modify the name of the output file
3) Specify the offset of the target data from GMT

Author: Drew Allan Loney, PhD PE
Email: drew.a.loney@usace.army.mil
Version: 1.0

It is requested that any errors found within the script be conveyed to the script author. These will be corrected
in a timely manner, with an updated script posted for the community's use.


This script is released under the MIT license:   
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated 
documentation files (the "Software"), to deal in the Software without restriction, including without limitation the 
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to 
permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE 
WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

   """

### Imports ###
import numpy as np
from scipy.interpolate import interp1d
import datetime, warnings
from multiprocessing import Pool, freeze_support
from itertools import repeat


### User specified values ###
s_input_filename = 'hydromet_KUKI_19730101_20160430.txt'			# Name of the input file
s_output_filename = '2017_07_27_ukiah_airport.hmt'					# Name of the output file
i_gmt_offset = -8     # hours


####################### No additional user input after this line #######################################################
def _local_value_constructor(o_timestep, dm_hmet_input):
    print('\t' + str(o_timestep))
    i_index = np.argwhere(np.logical_and(np.logical_and(o_timestep.year == dm_hmet_input[:, 0],
                                                        o_timestep.month == dm_hmet_input[:, 1]),
                                         np.logical_and(o_timestep.day == dm_hmet_input[:, 2],
                                                        o_timestep.hour == dm_hmet_input[:, 3]))).flatten()

    # Append the value into the array
    if len(i_index) == 0:
        # dm_values[i_entry_timestep, :] = np.ones(7) * np.NaN
        return np.NaN

    else:
        # dm_values[i_entry_timestep, :] = dm_hmet_input[i_index[0], 4:]
        # ia_values_indices.append(i_index[0])
        return i_index[0]

if __name__ == '__main__':
    freeze_support()
	
	### Define the time increment ###
	i_time_increment = 60 # minutes

    ### Create the OMP worker pool ###
    print('Opening the worker pool...')
    # Open the pool
    o_pool = Pool()

    ### Load the hmet file into memory ###
    print('Loading source HMET file...')
    dm_hmet_input = np.genfromtxt(s_input_filename, delimiter='\t', filling_values=np.NaN, missing_values=np.NaN)

    # Get the start and ending dates
    o_start_date_gage = datetime.datetime(year=int(dm_hmet_input[0,0]), month=int(dm_hmet_input[0,1]),
                                          day=int(dm_hmet_input[0,2]), hour=int(dm_hmet_input[0,3]))

    i_stop_index = len(dm_hmet_input) - 1
    o_stop_date_gage = datetime.datetime(year=int(dm_hmet_input[i_stop_index,0]), month=int(dm_hmet_input[i_stop_index,1]),
                                         day=int(dm_hmet_input[i_stop_index,2]), hour=int(dm_hmet_input[i_stop_index,3]))

    ### Fill the series between the start and stop dates ###
    # Create a timedelta between series values
    o_timedelta = datetime.timedelta(minutes=i_time_increment)

    # Calculate the number of timesteps needed to construct the series
    i_number_of_increments = ((o_stop_date_gage - o_start_date_gage).total_seconds() / 60.0) / i_time_increment

    ### Construct a timestep list ###
    print('Constructing the timestep list...')
    l_timesteps = [o_start_date_gage]

    o_running_date_gage = o_start_date_gage
    for i_entry_timestep in range(0, int(i_number_of_increments), 1):
        # Construct each of the timesteps
        o_running_date_gage += o_timedelta

        # Append into the list
        l_timesteps.append(o_running_date_gage)

    ### Create the value list ###
    print('Constructing the value matrix...')
    dm_values = np.zeros((int(i_number_of_increments+1), 7))
    ia_values_indices = []

    # Determine the indices in the expanded set
    ia_indices = np.array(list(o_pool.starmap(_local_value_constructor, zip(l_timesteps, repeat(dm_hmet_input)))))

    # Find indices that are valid
    ia_indices_valid = np.argwhere(~np.isnan(ia_indices)).flatten()

    # Place the data into the new matrix
    dm_values[ia_indices_valid.astype(int), :] = dm_hmet_input[ia_indices[ia_indices_valid].astype(int), 4:]

    # Find the indices that are invalid
    ia_indices_invalid = np.argwhere(np.isnan(ia_indices)).flatten()

    # Set invalid values as Nan
    dm_values[ia_indices_invalid.astype(int), :] = np.NaN

    # Promote the index list
    ia_values_indices = np.array(ia_indices[ia_indices_valid])

    ### Interpolate for the values ###
    print('Interpolating for missing values...')
    ia_spacing = np.arange(0, i_number_of_increments+1, 1)

    # Create the array to be filled
    dm_values_interpolated = np.zeros(np.shape(dm_values))

    # Fill with the GSSHA default values
    for i_entry in range(0, len(dm_values), 1):

        if np.all(np.isnan(dm_values[i_entry, :])):
            dm_values_interpolated[i_entry, :] = np.array([99.999, 999, 999, 999, 999, 9999.99, 9999.99])
        else:
            dm_values_interpolated[i_entry, :] = dm_values[i_entry, :]


    ### Correct the timesteps to local from GMT ###
    print('Correcting for GMT offset...')
    o_timedelta = datetime.timedelta(hours=i_gmt_offset)
    l_timesteps = [x + o_timedelta for x in l_timesteps]


    ### Add the date to the matrix ###
    print('Adding dates to the output matrix...')
    # Create the empty array of the appropriate length
    dm_values_interpolated = np.hstack((np.zeros((len(l_timesteps), 4)), dm_values_interpolated))

    for i_entry_row in range(0, len(l_timesteps), 1):
        dm_values_interpolated[i_entry_row, 0] = l_timesteps[i_entry_row].year
        dm_values_interpolated[i_entry_row, 1] = l_timesteps[i_entry_row].month
        dm_values_interpolated[i_entry_row, 2] = l_timesteps[i_entry_row].day
        dm_values_interpolated[i_entry_row, 3] = l_timesteps[i_entry_row].hour


    ### Write the file to disk ###
    print('Outputting file...')
    o_file = open(s_output_filename, 'w')

    for da_row in dm_values_interpolated:
        # Write the timestamp
        o_file.write(str(int(da_row[0])) + '\t' + str(int(da_row[1])) + '\t' + str(int(da_row[2])) + '\t' +
                     str(int(da_row[3])) + '\t')

        # Write the barometric pressure (inches Hg)
        if not np.isnan(da_row[4]) and da_row[4] !=  99.999:
            o_file.write(str(round(da_row[4], 3)) + '\t')
        else:
            o_file.write(str(99.999) + '\t')

        # Write the relative humidity (percent)
        if not np.isnan(da_row[5]) and da_row[5] !=  999:
            o_file.write(str(int(round(da_row[5], 0))) + '\t')
        else:
            o_file.write(str(999) + '\t')

        # Write the total sky cover (percent)
        if not np.isnan(da_row[6]) and da_row[6] !=  999:
            o_file.write(str(int(round(da_row[6], 0))) + '\t')
        else:
            o_file.write(str(999) + '\t')

        # Write the wind speed (knots)
        if not np.isnan(da_row[7]) and da_row[7] !=  999:
            o_file.write(str(int(round(da_row[7], 0))) + '\t')
        else:
            o_file.write(str(999) + '\t')

        # Write the dry bulb temp (degrees F)
        if not np.isnan(da_row[8]) and da_row[8] !=  999:
            o_file.write(str(int(round(da_row[8], 0))) + '\t')
        else:
            o_file.write(str(999) + '\t')

        # Write the direct radiation
        if not np.isnan(da_row[9]) and da_row[9] !=  9999.99:
            o_file.write(str(round(da_row[9], 1)) + '\t')
        else:
            o_file.write(str(round(9999.99, 2)) + '\t')

        # Write the global radiation
        if not np.isnan(da_row[10]) and da_row[10] != 9999.99:
            o_file.write(str(round(da_row[10], 1)) + '\n')
        else:
            o_file.write(str(round(9999.99, 2)) + '\n')

    # Close the file
    o_file.close()